# File generated from our OpenAPI spec by Stainless. See CONTRIBUTING.md for details.

from typing_extensions import Literal

from .._models import BaseModel

__all__ = ["ServerToolCaller"]


class ServerToolCaller(BaseModel):
    """Tool invocation generated by a server-side tool."""

    tool_id: str

    type: Literal["code_execution_20250825"]
