"""Auto-generated model profiles.

DO NOT EDIT THIS FILE MANUALLY.
This file is generated by the langchain-profiles CLI tool.

It contains data derived from the models.dev project.

Source: https://github.com/sst/models.dev
License: MIT License

To update these data, refer to the instructions here:

https://docs.langchain.com/oss/python/langchain/models#updating-or-overwriting-profile-data
"""

from typing import Any

_PROFILES: dict[str, dict[str, Any]] = {
    "gemini-embedding-001": {
        "max_input_tokens": 2048,
        "max_output_tokens": 3072,
        "text_inputs": True,
        "image_inputs": False,
        "audio_inputs": False,
        "video_inputs": False,
        "text_outputs": True,
        "image_outputs": False,
        "audio_outputs": False,
        "video_outputs": False,
        "reasoning_output": False,
        "tool_calling": False,
        "image_url_inputs": True,
        "pdf_inputs": True,
        "image_tool_message": True,
        "tool_choice": True,
    },
    "gemini-2.5-flash-lite-preview-09-2025": {
        "max_input_tokens": 1048576,
        "max_output_tokens": 65536,
        "text_inputs": True,
        "image_inputs": True,
        "audio_inputs": True,
        "pdf_inputs": True,
        "video_inputs": True,
        "text_outputs": True,
        "image_outputs": False,
        "audio_outputs": False,
        "video_outputs": False,
        "reasoning_output": True,
        "tool_calling": True,
        "structured_output": True,
        "image_url_inputs": True,
        "image_tool_message": True,
        "tool_choice": True,
    },
    "gemini-3.1-pro-preview-customtools": {
        "max_input_tokens": 1048576,
        "max_output_tokens": 65536,
        "text_inputs": True,
        "image_inputs": True,
        "audio_inputs": True,
        "pdf_inputs": True,
        "video_inputs": True,
        "text_outputs": True,
        "image_outputs": False,
        "audio_outputs": False,
        "video_outputs": False,
        "reasoning_output": True,
        "tool_calling": True,
        "structured_output": True,
        "image_url_inputs": True,
        "image_tool_message": True,
        "tool_choice": True,
    },
    "gemini-2.5-pro-preview-06-05": {
        "max_input_tokens": 1048576,
        "max_output_tokens": 65536,
        "text_inputs": True,
        "image_inputs": True,
        "audio_inputs": True,
        "pdf_inputs": True,
        "video_inputs": True,
        "text_outputs": True,
        "image_outputs": False,
        "audio_outputs": False,
        "video_outputs": False,
        "reasoning_output": True,
        "tool_calling": True,
        "structured_output": True,
        "image_url_inputs": True,
        "image_tool_message": True,
        "tool_choice": True,
    },
    "gemini-2.5-flash-preview-04-17": {
        "max_input_tokens": 1048576,
        "max_output_tokens": 65536,
        "text_inputs": True,
        "image_inputs": True,
        "audio_inputs": True,
        "pdf_inputs": True,
        "video_inputs": True,
        "text_outputs": True,
        "image_outputs": False,
        "audio_outputs": False,
        "video_outputs": False,
        "reasoning_output": True,
        "tool_calling": True,
        "image_url_inputs": True,
        "image_tool_message": True,
        "tool_choice": True,
    },
    "gemini-2.5-flash-preview-09-2025": {
        "max_input_tokens": 1048576,
        "max_output_tokens": 65536,
        "text_inputs": True,
        "image_inputs": True,
        "audio_inputs": True,
        "pdf_inputs": True,
        "video_inputs": True,
        "text_outputs": True,
        "image_outputs": False,
        "audio_outputs": False,
        "video_outputs": False,
        "reasoning_output": True,
        "tool_calling": True,
        "structured_output": True,
        "image_url_inputs": True,
        "image_tool_message": True,
        "tool_choice": True,
    },
    "gemini-2.5-pro-preview-05-06": {
        "max_input_tokens": 1048576,
        "max_output_tokens": 65536,
        "text_inputs": True,
        "image_inputs": True,
        "audio_inputs": True,
        "pdf_inputs": True,
        "video_inputs": True,
        "text_outputs": True,
        "image_outputs": False,
        "audio_outputs": False,
        "video_outputs": False,
        "reasoning_output": True,
        "tool_calling": True,
        "structured_output": True,
        "image_url_inputs": True,
        "image_tool_message": True,
        "tool_choice": True,
    },
    "gemini-2.5-flash-preview-05-20": {
        "max_input_tokens": 1048576,
        "max_output_tokens": 65536,
        "text_inputs": True,
        "image_inputs": True,
        "audio_inputs": True,
        "pdf_inputs": True,
        "video_inputs": True,
        "text_outputs": True,
        "image_outputs": False,
        "audio_outputs": False,
        "video_outputs": False,
        "reasoning_output": True,
        "tool_calling": True,
        "structured_output": True,
        "image_url_inputs": True,
        "image_tool_message": True,
        "tool_choice": True,
    },
    "gemini-2.5-flash": {
        "max_input_tokens": 1048576,
        "max_output_tokens": 65536,
        "text_inputs": True,
        "image_inputs": True,
        "audio_inputs": True,
        "pdf_inputs": True,
        "video_inputs": True,
        "text_outputs": True,
        "image_outputs": False,
        "audio_outputs": False,
        "video_outputs": False,
        "reasoning_output": True,
        "tool_calling": True,
        "structured_output": True,
        "image_url_inputs": True,
        "image_tool_message": True,
        "tool_choice": True,
    },
    "gemini-live-2.5-flash": {
        "max_input_tokens": 128000,
        "max_output_tokens": 8000,
        "text_inputs": True,
        "image_inputs": True,
        "audio_inputs": True,
        "video_inputs": True,
        "text_outputs": True,
        "image_outputs": False,
        "audio_outputs": True,
        "video_outputs": False,
        "reasoning_output": True,
        "tool_calling": True,
        "image_url_inputs": True,
        "pdf_inputs": True,
        "image_tool_message": True,
        "tool_choice": True,
    },
    "gemini-3-flash-preview": {
        "max_input_tokens": 1048576,
        "max_output_tokens": 65536,
        "text_inputs": True,
        "image_inputs": True,
        "audio_inputs": True,
        "pdf_inputs": True,
        "video_inputs": True,
        "text_outputs": True,
        "image_outputs": False,
        "audio_outputs": False,
        "video_outputs": False,
        "reasoning_output": True,
        "tool_calling": True,
        "structured_output": True,
        "image_url_inputs": True,
        "image_tool_message": True,
        "tool_choice": True,
    },
    "gemini-live-2.5-flash-preview-native-audio": {
        "max_input_tokens": 131072,
        "max_output_tokens": 65536,
        "text_inputs": True,
        "image_inputs": False,
        "audio_inputs": True,
        "video_inputs": True,
        "text_outputs": True,
        "image_outputs": False,
        "audio_outputs": True,
        "video_outputs": False,
        "reasoning_output": True,
        "tool_calling": True,
        "image_url_inputs": True,
        "pdf_inputs": True,
        "image_tool_message": True,
        "tool_choice": True,
    },
    "gemini-2.5-flash-lite": {
        "max_input_tokens": 1048576,
        "max_output_tokens": 65536,
        "text_inputs": True,
        "image_inputs": True,
        "audio_inputs": True,
        "pdf_inputs": True,
        "video_inputs": True,
        "text_outputs": True,
        "image_outputs": False,
        "audio_outputs": False,
        "video_outputs": False,
        "reasoning_output": True,
        "tool_calling": True,
        "structured_output": True,
        "image_url_inputs": True,
        "image_tool_message": True,
        "tool_choice": True,
    },
    "gemini-2.5-flash-preview-tts": {
        "max_input_tokens": 8000,
        "max_output_tokens": 16000,
        "text_inputs": True,
        "image_inputs": False,
        "audio_inputs": False,
        "video_inputs": False,
        "text_outputs": False,
        "image_outputs": False,
        "audio_outputs": True,
        "video_outputs": False,
        "reasoning_output": False,
        "tool_calling": False,
        "image_url_inputs": True,
        "pdf_inputs": True,
        "image_tool_message": True,
        "tool_choice": True,
    },
    "gemini-3.1-pro-preview": {
        "max_input_tokens": 1048576,
        "max_output_tokens": 65536,
        "text_inputs": True,
        "image_inputs": True,
        "audio_inputs": True,
        "pdf_inputs": True,
        "video_inputs": True,
        "text_outputs": True,
        "image_outputs": False,
        "audio_outputs": False,
        "video_outputs": False,
        "reasoning_output": True,
        "tool_calling": True,
        "structured_output": True,
        "image_url_inputs": True,
        "image_tool_message": True,
        "tool_choice": True,
    },
    "gemini-flash-latest": {
        "max_input_tokens": 1048576,
        "max_output_tokens": 65536,
        "text_inputs": True,
        "image_inputs": True,
        "audio_inputs": True,
        "pdf_inputs": True,
        "video_inputs": True,
        "text_outputs": True,
        "image_outputs": False,
        "audio_outputs": False,
        "video_outputs": False,
        "reasoning_output": True,
        "tool_calling": True,
        "structured_output": True,
        "image_url_inputs": True,
        "image_tool_message": True,
        "tool_choice": True,
    },
    "gemini-2.5-flash-lite-preview-06-17": {
        "max_input_tokens": 1048576,
        "max_output_tokens": 65536,
        "text_inputs": True,
        "image_inputs": True,
        "audio_inputs": True,
        "pdf_inputs": True,
        "video_inputs": True,
        "text_outputs": True,
        "image_outputs": False,
        "audio_outputs": False,
        "video_outputs": False,
        "reasoning_output": True,
        "tool_calling": True,
        "image_url_inputs": True,
        "image_tool_message": True,
        "tool_choice": True,
    },
    "gemini-2.5-flash-image": {
        "max_input_tokens": 32768,
        "max_output_tokens": 32768,
        "text_inputs": True,
        "image_inputs": True,
        "audio_inputs": False,
        "video_inputs": False,
        "text_outputs": True,
        "image_outputs": True,
        "audio_outputs": False,
        "video_outputs": False,
        "reasoning_output": True,
        "tool_calling": False,
        "image_url_inputs": True,
        "pdf_inputs": True,
        "image_tool_message": True,
        "tool_choice": True,
    },
    "gemini-2.5-pro-preview-tts": {
        "max_input_tokens": 8000,
        "max_output_tokens": 16000,
        "text_inputs": True,
        "image_inputs": False,
        "audio_inputs": False,
        "video_inputs": False,
        "text_outputs": False,
        "image_outputs": False,
        "audio_outputs": True,
        "video_outputs": False,
        "reasoning_output": False,
        "tool_calling": False,
        "image_url_inputs": True,
        "pdf_inputs": True,
        "image_tool_message": True,
        "tool_choice": True,
    },
    "gemini-2.5-flash-image-preview": {
        "max_input_tokens": 32768,
        "max_output_tokens": 32768,
        "text_inputs": True,
        "image_inputs": True,
        "audio_inputs": False,
        "video_inputs": False,
        "text_outputs": True,
        "image_outputs": True,
        "audio_outputs": False,
        "video_outputs": False,
        "reasoning_output": True,
        "tool_calling": False,
        "image_url_inputs": True,
        "pdf_inputs": True,
        "image_tool_message": True,
        "tool_choice": True,
    },
    "gemini-1.5-flash-8b": {
        "max_input_tokens": 1000000,
        "max_output_tokens": 8192,
        "text_inputs": True,
        "image_inputs": True,
        "audio_inputs": True,
        "video_inputs": True,
        "text_outputs": True,
        "image_outputs": False,
        "audio_outputs": False,
        "video_outputs": False,
        "reasoning_output": False,
        "tool_calling": True,
        "image_url_inputs": True,
        "pdf_inputs": True,
        "image_tool_message": True,
        "tool_choice": True,
    },
    "gemini-3-pro-preview": {
        "max_input_tokens": 1000000,
        "max_output_tokens": 64000,
        "text_inputs": True,
        "image_inputs": True,
        "audio_inputs": True,
        "pdf_inputs": True,
        "video_inputs": True,
        "text_outputs": True,
        "image_outputs": False,
        "audio_outputs": False,
        "video_outputs": False,
        "reasoning_output": True,
        "tool_calling": True,
        "structured_output": True,
        "image_url_inputs": True,
        "image_tool_message": True,
        "tool_choice": True,
    },
    "gemini-2.0-flash-lite": {
        "max_input_tokens": 1048576,
        "max_output_tokens": 8192,
        "text_inputs": True,
        "image_inputs": True,
        "audio_inputs": True,
        "pdf_inputs": True,
        "video_inputs": True,
        "text_outputs": True,
        "image_outputs": False,
        "audio_outputs": False,
        "video_outputs": False,
        "reasoning_output": False,
        "tool_calling": True,
        "structured_output": True,
        "image_url_inputs": True,
        "image_tool_message": True,
        "tool_choice": True,
    },
    "gemini-1.5-flash": {
        "max_input_tokens": 1000000,
        "max_output_tokens": 8192,
        "text_inputs": True,
        "image_inputs": True,
        "audio_inputs": True,
        "video_inputs": True,
        "text_outputs": True,
        "image_outputs": False,
        "audio_outputs": False,
        "video_outputs": False,
        "reasoning_output": False,
        "tool_calling": True,
        "image_url_inputs": True,
        "pdf_inputs": True,
        "image_tool_message": True,
        "tool_choice": True,
    },
    "gemini-flash-lite-latest": {
        "max_input_tokens": 1048576,
        "max_output_tokens": 65536,
        "text_inputs": True,
        "image_inputs": True,
        "audio_inputs": True,
        "pdf_inputs": True,
        "video_inputs": True,
        "text_outputs": True,
        "image_outputs": False,
        "audio_outputs": False,
        "video_outputs": False,
        "reasoning_output": True,
        "tool_calling": True,
        "structured_output": True,
        "image_url_inputs": True,
        "image_tool_message": True,
        "tool_choice": True,
    },
    "gemini-2.5-pro": {
        "max_input_tokens": 1048576,
        "max_output_tokens": 65536,
        "text_inputs": True,
        "image_inputs": True,
        "audio_inputs": True,
        "pdf_inputs": True,
        "video_inputs": True,
        "text_outputs": True,
        "image_outputs": False,
        "audio_outputs": False,
        "video_outputs": False,
        "reasoning_output": True,
        "tool_calling": True,
        "structured_output": True,
        "image_url_inputs": True,
        "image_tool_message": True,
        "tool_choice": True,
    },
    "gemini-2.0-flash": {
        "max_input_tokens": 1048576,
        "max_output_tokens": 8192,
        "text_inputs": True,
        "image_inputs": True,
        "audio_inputs": True,
        "pdf_inputs": True,
        "video_inputs": True,
        "text_outputs": True,
        "image_outputs": False,
        "audio_outputs": False,
        "video_outputs": False,
        "reasoning_output": False,
        "tool_calling": True,
        "structured_output": True,
        "image_url_inputs": True,
        "image_tool_message": True,
        "tool_choice": True,
    },
    "gemini-1.5-pro": {
        "max_input_tokens": 1000000,
        "max_output_tokens": 8192,
        "text_inputs": True,
        "image_inputs": True,
        "audio_inputs": True,
        "video_inputs": True,
        "text_outputs": True,
        "image_outputs": False,
        "audio_outputs": False,
        "video_outputs": False,
        "reasoning_output": False,
        "tool_calling": True,
        "image_url_inputs": True,
        "pdf_inputs": True,
        "image_tool_message": True,
        "tool_choice": True,
    },
}
