# File generated from our OpenAPI spec by Stainless. See CONTRIBUTING.md for details.

from __future__ import annotations

from typing_extensions import Literal, Required, TypedDict

__all__ = ["ServerToolCallerParam"]


class ServerToolCallerParam(TypedDict, total=False):
    """Tool invocation generated by a server-side tool."""

    tool_id: Required[str]

    type: Required[Literal["code_execution_20250825"]]
